# Copyright (c) 2022 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'ROOT',
    'COLORS',
    'COLOR_DICT',
    'thermochemical_equilibrium',
    'thermo_eval',
    'stoich_matrix',
    'read_elemental',
    'set_element_abundance',
    'de_aliasing',
    'resolve_sources',
    'resolve_colors',
    'plot_vmr',
    'write_file',
    'read_file',
]

import itertools
import os
from pathlib import Path
import sys
import warnings

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

ROOT = str(Path(__file__).parents[2]) + os.path.sep

sys.path.append(f'{ROOT}chemcat/lib')
import _thermo as nr


# A long list of colors:
COLORS = [
    'royalblue', 'darkorange', 'red', 'darkgreen', 'magenta',
    'blue', 'limegreen', 'gold', 'dimgray', 'navy',
    'deepskyblue', 'silver', 'black', 'olive', 'chocolate',
    'skyblue', 'darkviolet', 'greenyellow', 'pink', 'coral',
    'darkcyan', 'rosybrown', 'cornflowerblue', 'mediumvioletred', 'maroon',
    'darkgoldenrod', 'darkkhaki', 'hotpink', 'darkslateblue', 'lightgreen',
    'yellowgreen', 'seagreen', 'yellow', 'slateblue', 'sienna',
    'peachpuff', 'orangered', 'goldenrod', 'brown', 'khaki',
    # Unique line
    'saddlebrown', 'mediumseagreen', 'darksalmon', 'cadetblue',
    'mediumaquamarine', 'darkslategray', 'lightsteelblue', 'indigo',
    'lightcoral', 'lightslategray', 'lawngreen', 'lightblue',
    'darkseagreen', 'sandybrown', 'tan', 'slategray',
    'steelblue', 'wheat', 'mediumslateblue', 'mediumorchid',
    'cyan', 'springgreen', 'lime', 'dodgerblue', 'deeppink',
    'mediumblue', 'green', 'tomato', 'crimson', 'palegoldenrod',
    'lightsalmon', 'forestgreen', 'orchid', 'turquoise', 'darkolivegreen',
    'lightseagreen', 'violet', 'salmon', 'indianred', 'rebeccapurple',
    'peru', 'darkturquoise', 'lightskyblue', 'plum', 'aquamarine',
    'mediumspringgreen', 'orange', 'purple', 'midnightblue', 'darkgray',
    'darkorchid', 'blueviolet', 'teal', 'darkmagenta', 'palevioletred',
    'firebrick', 'mediumpurple', 'gainsboro',
]

COLOR_DICT = {
    'H': 'blue',
    'H2': 'deepskyblue',
    'He': 'olive',
    # Carbons:
    'C': 'coral',
    'CH4': 'darkorange',
    'CO': 'limegreen',
    'CO2': 'red',
    'HCN': 'dimgray',
    'C2H2': 'pink',
    'C2H4': 'deeppink',
    # Nitrogens:
    'N': 'darkviolet',
    'NH3': 'magenta',
    'N2': 'gold',
    # Oxygens:
    'O': 'greenyellow',
    'H2O': 'navy',
    'OH': 'darkkhaki',
    # Silicons:
    'Si': 'lightslategray',
    'SiO': 'darkturquoise',
    'SiH4': 'mediumvioletred',
    # Alkali:
    'Na': 'silver',
    '(NaCl)2': 'maroon',
    '(NaOH)2': 'hotpink',
    'NaCl': 'rosybrown',
    'K': 'black',
    '(KCl)2': 'chocolate',
    '(KOH)2': 'darkslateblue',
    'KOH': 'lightgreen',
    'KCl': 'darksalmon',
    # Sulfurs:
    'S': 'cornflowerblue',
    'H2S': 'darkgoldenrod',
    'HS': 'yellowgreen',
    'SO': 'mediumseagreen',
    'SO2': 'skyblue',
    # Aluminum:
    'Al': 'khaki',
    'AlOH': 'steelblue',
    'Al2O': 'seagreen',
    'OAlOH': 'tomato',
    'Ca': 'orange',
    'Ca(OH)2': 'indigo',
    'e': 'darkgreen',
    # Heavy metals:
    'Ti': 'crimson',
    'TiO': 'brown',
    'TiO2': 'indianred',
    'VO': 'aquamarine',
    'VO2': 'mediumaquamarine',
    'V': 'darkcyan',
    'Mg': 'sandybrown',
    'MgH': 'lawngreen',
    'Mg(OH)2': 'orangered',
    'Fe': 'royalblue',
    'FeH': 'wheat',
    'Fe(OH)2': 'tan',
    'F': 'yellow',
    'OAlF2': 'sienna',
    'TiF3': 'saddlebrown',
    'AlF': 'orange',
    'HF': 'lightblue',
    'MnH': 'lime',
    'Mn': 'rebeccapurple',
    'PN': 'palegoldenrod',
    'P': 'peachpuff',
    '(P2O3)2': 'cadetblue',
}


def thermochemical_equilibrium(
        pressure, temperature, element_rel_abundance, stoich_vals,
        gibbs_funcs, tolx=2.22e-16, tolf=2.22e-16,
    ):
    """
    Low-level function to compute thermochemical equilibrium for the
    given chemical network at the specified temperature--pressure
    profile.

    Parameters
    ----------
    pressure: 1D float array
        Pressure profile (bar).
    temperature: 1D float array
        Temperature profile (Kelvin).
    element_rel_abundance: 1D float array
        Elemental abundances (relative to H=1.0).
    stoich_vals: 2D float array
        Species stoichiometric values for CHON.
    gibbs_funcs: 1D iterable of callable functions
        Functions that return the Gibbs free energy (divided by RT)
        for each species in the network.
    tolx: float
        Relative error desired for convergence in the sum of squares.
    tolf: float
        Relative error desired for convergence in the approximate solution.

    Returns
    -------
    vmr: 2D float array
        Species volume mixing ratios in thermochemical equilibrium
        of shape [nlayers, nspecies].
    """
    nlayers = len(pressure)
    nspecies, nelements = np.shape(stoich_vals)

    # Target elemental fractions (and charge) for conservation:
    b0 = element_rel_abundance / np.sum(element_rel_abundance)
    # Total elemental abundance as first guess for total species abundance:
    total_abundance = np.sum(b0)

    is_atom = element_rel_abundance > 0.0
    # Maximum abundance reachable by each species
    # (i.e., species takes all available atoms)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        max_abundances = np.array([
            np.nanmin((b0/is_atom)[stoich>0] / stoich[stoich>0])
            for stoich in stoich_vals
        ])

    total_natoms = np.sum(stoich_vals*is_atom, axis=1)
    electron_index = total_natoms == 0
    max_abundances[electron_index] = total_abundance

    # Initial guess for abundances of species, gets modified in-place
    # with best-fit values by nr.gibbs_energy_minimizer()
    abundances = np.copy(max_abundances)

    # Number of conservation equations to solve with Newton-Raphson
    nequations = nelements + 1
    pilag = np.zeros(nelements)  # pi lagrange multiplier
    x = np.zeros(nequations)

    vmr = np.zeros((nlayers, nspecies))
    mu = np.zeros(nspecies)  # chemical potential/RT
    h_ts = thermo_eval(temperature, gibbs_funcs).T + np.log(pressure)
    delta_ln_vmr = np.zeros(nspecies)

    # Compute thermochemical equilibrium abundances at each layer:
    # (Go down the layers and then sweep back up in case the first run
    # didn't get the global minimum)
    for i in itertools.chain(range(nlayers), reversed(range(nlayers))):
        abundances[abundances <= 0] = 1e-300
        exit_status = nr.gibbs_energy_minimizer(
            nspecies, nequations, stoich_vals, b0,
            temperature[i], h_ts[:,i], pilag,
            abundances, max_abundances, total_abundance,
            mu, x, delta_ln_vmr, tolx, tolf,
        )

        if exit_status == 1:
            print(f"Gibbs minimization failed at layer {i}")
        vmr[i] = abundances / np.sum(abundances[~electron_index])

    return vmr


def thermo_eval(temperature, thermo_funcs):
    r"""
    Low-level function to compute the thermochemical property
    specified by thermo_func at at the requested temperature(s).
    These can be, e.g., the heat_capacity or gibbs_free_energy
    functions returned by setup_network().

    Normally you want to use this function via the heat_capacity()
    and gibbs_free_energy() methods of the chemcat.Network() object.

    Parameters
    ----------
    temperature: float or 1D float iterable
        Temperature (Kelvin).
    thermo_funcs: 1D iterable of callable functions
        Functions that return the thermochemical property.

    Returns
    -------
    thermo_prop: 1D or 2D float array
        The provided thermochemical property evaluated at the requested
        temperature(s).
        The shape of the output depends on the shape of the
        temperature input.

    Examples
    --------
    >>> import chemcat as cat
    >>> import chemcat.janaf as janaf
    >>> import chemcat.utils as u
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np

    >>> molecules = (
    >>>     'H2O CH4 CO CO2 NH3 N2 H2 HCN OH C2H2 C2H4 H He C N O').split()
    >>> janaf_data = janaf.setup_network(molecules)
    >>> species = janaf_data[0]
    >>> heat_funcs = janaf_data[1]
    >>> gibbs_funcs = janaf_data[2]

    >>> temperature = 1500.0
    >>> temperatures = np.arange(100.0, 4501.0, 10)
    >>> cp1 = u.thermo_eval(temperature, heat_funcs)
    >>> cp2 = u.thermo_eval(temperatures, heat_funcs)
    >>> gibbs = u.thermo_eval(temperatures, gibbs_funcs)

    >>> nspecies = len(species)
    >>> plt.figure('Heat capacity, Gibbs free energy', (8.5, 4.5))
    >>> plt.clf()
    >>> plt.subplot(121)
    >>> for j in range(nspecies):
    >>>     label = species[j]
    >>>     plt.plot(
    >>>         temperatures, cp2[:,j], label=label, c=u.COLOR_DICT[label],
    >>>     )
    >>> plt.xlim(np.amin(temperatures), np.amax(temperatures))
    >>> plt.plot(np.tile(temperature,nspecies), cp1, 'ob', ms=4, zorder=-1)
    >>> plt.xlabel('Temperature (K)')
    >>> plt.ylabel('Heat capacity / R')

    >>> plt.subplot(122)
    >>> for j in range(nspecies):
    >>>     label = species[j]
    >>>     plt.plot(
    >>>         temperatures, gibbs[:,j], label=label, c=u.COLOR_DICT[label],
    >>>     )
    >>> plt.xlim(np.amin(temperatures), np.amax(temperatures))
    >>> plt.legend(loc='upper right', fontsize=8)
    >>> plt.xlabel('Temperature (K)')
    >>> plt.ylabel('Gibbs free energy / RT')
    >>> plt.tight_layout()
    """
    temp = np.atleast_1d(temperature)
    ntemp = np.shape(temp)[0]
    nspecies = len(thermo_funcs)
    thermo_prop = np.zeros((ntemp, nspecies))
    for j in range(nspecies):
        thermo_prop[:,j] = thermo_funcs[j](temp)
    if np.shape(temperature) == ():
        return np.squeeze(thermo_prop)
    return thermo_prop


def stoich_matrix(stoich_data):
    r"""
    Compute matrix of stoichiometric values for the given stoichiometric
    data for a network of species.

    Parameters
    ----------
    stoich_data: List of dictionaries
        Stoichiometric data (as dictionary of element-value pairs) for
        a list of species.

    Returns
    -------
    elements: 1D string array
        Elements for this chemical network.
    stoich_vals: 2D integer array
        Array containing the stoichiometric values for the
        requested species sorted according to the species and elements
        arrays.

    Examples
    --------
    >>> import chemcat.utils as u
    >>> stoich_data = [
    >>>     {'H': 2.0, 'O': 1.0},
    >>>     {'C': 1.0, 'H': 4.0},
    >>>     {'C': 1.0, 'O': 2.0},
    >>>     {'H': 2.0},
    >>>     {'H': 1.0},
    >>>     {'He': 1.0},
    >>> ]
    >>> elements, stoich_matrix = u.stoich_matrix(stoich_data)
    >>> print(elements, stoich_matrix, sep='\n')
    ['C' 'H' 'He' 'O']
    [[0 2 0 1]
     [1 4 0 0]
     [1 0 0 2]
     [0 2 0 0]
     [0 1 0 0]
     [0 0 1 0]]
    """
    elements = []
    for s in stoich_data:
        elements += list(s.keys())
    elements = sorted(set(elements))

    nspecies = len(stoich_data)
    nelements = len(elements)
    stoich_vals = np.zeros((nspecies, nelements), int)
    for i in range(nspecies):
        for key,val in stoich_data[i].items():
            j = elements.index(key)
            stoich_vals[i,j] = val
    elements = np.array(elements)
    return elements, stoich_vals


def read_elemental(element_file):
    """
    Extract elemental abundances from a file (defaulted to a solar
    elemental abundance file from Asplund et al. 2021).

    Inputs
    ------
    element_file: String
        Path to a file containing a list of elements (second column)
        and their relative abundances in log10 scale relative to H=12.0
        (third column).

    Returns
    -------
    elements: 1D string array
        The list of elements.
    dex_abundances: 1D float array
        The elemental abundances in dex units relative to H=12.0.

    Examples
    --------
    >>> import chemcat.utils as u

    >>> element_file = f'{u.ROOT}chemcat/data/asplund_2021_solar_abundances.dat'
    >>> elements, dex = u.read_elemental(element_file)
    >>> for e in 'H He C N O'.split():
    >>>     print(f'{e:2}:  {dex[elements==e][0]:6.3f}')
    H :  12.000
    He:  10.914
    C :   8.460
    N :   7.830
    O :   8.690
    """
    elements, dex = np.loadtxt(
        element_file, dtype=str, comments='#', usecols=(1,2), unpack=True)
    dex_abundances = np.array(dex, float)
    return elements, dex_abundances


def set_element_abundance(
        elements, base_composition, base_dex_abundances,
        metallicity=0.0, e_abundances={}, e_scale={}, e_ratio={},
    ):
    """
    Set an elemental composition by scaling metals and custom atomic species.

    Parameters
    ----------
    elements: 1D string array
        List of elements to return their abundances.
    base_composition: 1D float array
        List of all possible elements.
    base_dex_abundances: 1D float iterable
        The elemental base abundances in dex units relative to H=12.0.
    metallicity: Float
        Scaling factor for all elemental species except H and He.
        Dex units relative to the sun (e.g., solar is metallicity=0.0,
        10x solar is metallicity=1.0).
    e_abundances: Dictionary of element-abundance pairs
        Set custom elemental abundances.
        The dict contains the name of the element and their custom
        abundance in dex units relative to H=12.0.
        These values (if any) override metallicity.
    e_scale: Dictionary of element-scaling pairs
        Set custom elemental abundances by scaling from its solar value.
        The dict contains the name of the element and their custom
        scaling factor in dex units, e.g., for 2x solar carbon set
        e_scale = {'C': np.log10(2.0)}.
        This argument modifies the abundances on top of any custom
        metallicity and e_abundances.
    e_ratio: Dictionary of element-ratio pairs
        Set custom elemental abundances by scaling relative to another
        element.
        The dict contains the pair of elements joined by an underscore
        and their ratio, e.g., for a C/O ratio of 0.8 set
        e_ratio = {'C_O': 0.8}.
        These values scale on top of any custom metallicity,
        e_abundances, and e_scale.

    Returns
    -------
    elemental_abundances: 1D float array
        Elemental volume mixing ratios relative to H=1.0.

    Examples
    --------
    >>> import chemcat as cat
    >>> import chemcat.utils as u

    >>> element_file = f'{u.ROOT}chemcat/data/asplund_2021_solar_abundances.dat'
    >>> sun_elements, sun_dex = u.read_elemental(element_file)
    >>> elements = 'H He C N O'.split()

    >>> solar = u.set_element_abundance(
    >>>     elements, sun_elements, sun_dex,
    >>> )

    >>> # Set custom metallicity to [M/H] = 0.5:
    >>> abund = u.set_element_abundance(
    >>>     elements, sun_elements, sun_dex, metallicity=0.5,
    >>> )
    >>> for e,q in zip(elements, abund):
    >>>     print(f'{e:3} {q:.3e}')
    H   1.000e+00
    He  8.204e-02
    C   9.120e-04
    N   2.138e-04
    O   1.549e-03

    >>> # Custom carbon abundance by direct value (dex):
    >>> abund = u.set_element_abundance(
    >>>     elements, sun_elements, sun_dex, e_abundances={'C': 8.8},
    >>> )
    >>> for e,q in zip(elements, abund):
    >>>     print(f'{e:3} {q:.3e}')
    H   1.000e+00
    He  8.204e-02
    C   6.310e-04
    N   6.761e-05
    O   4.898e-04

    >>> # Custom carbon abundance by scaling to 2x its solar value:
    >>> abund = u.set_element_abundance(
    >>>     elements, sun_elements, sun_dex, e_scale={'C': np.log10(2)},
    >>> )
    >>> for e,q in zip(elements, abund):
    >>>     print(f'{e:3} {q:.3e}')
    H   1.000e+00
    He  8.204e-02
    C   5.768e-04
    N   6.761e-05
    O   4.898e-04

    >>> # Custom carbon abundance by scaling to C/O = 0.8:
    >>> abund = u.set_element_abundance(
    >>>     elements, sun_elements, sun_dex, e_ratio={'C_O': 0.8},
    >>> )
    >>> for e,q in zip(elements, abund):
    >>>     print(f'{e:3} {q:.3e}')
    H   1.000e+00
    He  8.204e-02
    C   3.918e-04
    N   6.761e-05
    O   4.898e-04
    """
    nelements = len(elements)
    elemental_abundances = np.zeros(nelements)
    for j in range(nelements):
        if elements[j] == 'e':
            continue
        idx = list(base_composition).index(elements[j])
        elemental_abundances[j] = base_dex_abundances[idx]

    # Scale the metals' abundances:
    imetals = np.isin(elements, 'H He D'.split(), invert=True)
    elemental_abundances[imetals] += metallicity
    # Set custom elemental abundances:
    for element, abundance in e_abundances.items():
        elemental_abundances[np.array(elements) == element] = abundance

    # Scale custom elemental abundances (additive to metallicity):
    for element, fscale in e_scale.items():
        elemental_abundances[np.array(elements) == element] += fscale

    # Set custom elemental ratios:
    for element, ratio in e_ratio.items():
        element1, element2 = element.split('_')
        idx1 = np.array(elements) == element1
        idx2 = np.array(elements) == element2
        log_ratio = np.log10(ratio)
        elemental_abundances[idx1] = elemental_abundances[idx2] + log_ratio

    # Convert elemental log VMR (relative to H=12.0) to VMR (rel. to H=1.0):
    elemental_abundances = 10**(elemental_abundances-12.0)
    elemental_abundances[np.array(elements) == 'e'] = 0.0
    return elemental_abundances


def de_aliasing(input_species, sources):
    """
    Get the right species names as given in the selected database.

    Parameters
    ----------
    input_species: List of strings
        List of species names.
    sources: String or 1D iterable of strings
        The desired database sources.

    Returns
    -------
    output_species: List of strings
        Species names with aliases replaced with the names
        as given in source database.

    Examples
    --------
    >>> import chemcat.utils as u
    >>> input_species = ['H2O', 'C2H2', 'HO2', 'CO']
    >>> sources = 'janaf'
    >>> output_species = u.de_aliasing(input_species, sources)
    >>> print(output_species)
    ['H2O', 'C2H2', 'HOO', 'CO']

    >>> sources = 'cea'
    >>> output_species = u.de_aliasing(input_species, sources)
    >>> print(output_species)
    ['H2O', 'C2H2,acetylene', 'HO2', 'CO']
    """
    if isinstance(sources, str):
        sources = [sources]

    # Get lists of species names aliases:
    whites = f'{ROOT}chemcat/data/white_pages.txt'
    aliases = []
    for line in open(whites, 'r'):
        if line.startswith('#'):
            continue
        aliases.append(line.split())
    all_aliases = np.concatenate(aliases)

    source_index = {
        'janaf': 0,
        'cea': 1,
    }

    output_species = []
    for species in input_species:
        if species not in all_aliases:
            output_species.append(species)
            continue
        # Find set of aliases:
        for alias in aliases:
            if species in alias:
                break
        # Search source in priority order:
        for source in sources:
            alias_name = alias[source_index[source]]
            if alias_name != 'None':
                output_species.append(alias_name)
                break
        else:
            output_species.append(species)
    return output_species


def resolve_sources(species, sources):
    r"""
    For each species in input, assign the right database proritizing
    by the order in sources.
    Parameters
    ----------
    species: 1D interable of strings
        Species to assign a database source.
    sources: 1D iterable of strings
        List of database sources in order of priority.
    Returns
    -------
    source_names: 1D array of strings
        Array with the assigned database to each species.
        If none found, leave value as None.
    Examples
    --------
    >>> import chemcat.utils as u
    >>> species = 'H2O CO (KOH)2 HO2'.split()
    >>> # Prioritize JANAF:
    >>> sources1 = u.resolve_sources(species, sources=['janaf', 'cea'])
    >>> # Prioritize CEA:
    >>> sources2 = u.resolve_sources(species, sources=['cea', 'janaf'])
    >>> # CEA exclusively:
    >>> sources3 = u.resolve_sources(species, sources=['cea'])
    >>> print(sources1, sources2, sources3, sep='\n')
    ['janaf' 'janaf' 'janaf' 'cea']
    ['cea' 'cea' 'janaf' 'cea']
    ['cea' 'cea' None 'cea']
    """
    # Need to import here to avoid circular import error in Python 3.6:
    from .. import janaf
    from .. import cea

    if isinstance(sources, str):
        sources = [sources]

    nspecies = len(species)
    source_names = np.tile(None, nspecies)

    for source in sources:
        if source == 'janaf':
            in_janaf = janaf.is_in(species)
            is_missing = [name is None for name in source_names]
            source_names[is_missing & in_janaf] = 'janaf'

        elif source == 'cea':
            in_cea = cea.is_in(species)
            is_missing = [name is None for name in source_names]
            source_names[is_missing & in_cea] = 'cea'

    return source_names


def resolve_colors(species, color_dict=None, color_list=None):
    """
    Assign a color for each input neutral species (ions will have
    the same color of a neutral form of the same species).

    Parameters
    ----------
    species: 1D string iterable
        The species that need to be assigned a color.
    color_dict: dict
        A dict with predefined colors for species.
        It does not need to contain a value for all input species.
        Defaulted to u.COLOR_DICT.
    color_list: 1D string iterable
        A list of color names to assing to the species that have not
        been assigned via color_dict.
        If there are more species than colors, then cycle over this
        list.
        Defaulted to u.COLORS.

    Returns
    -------
    colors: Dict
        Dict assigning a color to each of the (neutral) species.

    Examples
    --------
    >>> import chemcat.utils as u
    >>> species = 'H He C H2 CH4 CO CO2 e- H+ H- H3+'.split()
    >>> colors = u.resolve_colors(species)

    >>> print(colors)
    {'H': 'blue',
     'He': 'olive',
     'C': 'coral',
     'H2': 'deepskyblue',
     'CH4': 'darkorange',
     'CO': 'limegreen',
     'CO2': 'red',
     'e': 'darkgreen',
     'H3': 'royalblue'}
    """
    if color_list is not None and color_dict is None:
        color_dict = {}

    if color_dict is None:
        color_dict = COLOR_DICT
    if color_list is None:
        color_list = COLORS

    # Get all neutral species (plus electron):
    neutral_species = []
    for spec in species:
        neutral = spec.replace('-','').replace('+','')
        if neutral not in neutral_species:
            neutral_species.append(neutral)

    dicted_colors = [
        color_dict[spec]
        for spec in neutral_species
        if spec in color_dict
    ]
    remaining_colors = [
        color for color in color_list
        if color not in dicted_colors
    ]
    # Put dict_colors at the end, create an infinite cycle:
    colors_cycle = itertools.cycle(remaining_colors+dicted_colors)

    # Assign colors to each neutral species:
    # Species listed in dict:
    colors = {
        spec: color_dict[spec]
        for spec in neutral_species
        if spec in color_dict
    }
    # Remaining species:
    other_colors = {
        spec: next(colors_cycle)
        for spec in neutral_species
        if spec not in colors
    }
    colors.update(other_colors)
    return colors


def plot_vmr(
        pressure, vmr, species,
        colors=None, vmr_range=None, fignum=320, title=None,
        fontsize=14, linewidth=2.0, rect=None, axis=None,
        savefig=None,
    ):
    """
    Plot VMRs vs pressure.

    Parameters
    ----------
    pressure: 1D float iterable
        pressure array in bars.
    vmr: 2D float array
        Volume mixing ratios of shape [nlayers, nspecies].
    species: 1D string iterable
        Names of the species in vmr.
    colors: 1D iterable of strings
        Color names to assign (sequentially) to the species.
        If None, default to chemcat.utils.COLOR_DICT values.
        Note that different ionic variations of a same species
        (e.g., H, H+, H-) are assigned a same color, but differ
        in line style.
    vmr_range: 1D float iterable
        The plotting boundaries along the vmr axis.
    fignum: integer or string
        The identifier of the figure.
    title: Syting
        Title for the figure.
    fontsize: Float
        Font size for labels texts. Legend texts will be fontsize-5.
    linewidth: Float
        Width of VMR lines.
    rect: 4-element float iterable
        Axis position (left, bottom, right, top).
        Note that legend will be placed to the right of this rect.
    axis: AxesSubplot instance
        Axis where to draw the VMRs. If not None, overrides fignum.
    savefig: String
        If not None, file name where to save the figure.

    Returns
    -------
    ax: AxesSubplot instance
        The matplotlib Axes of the figure.

    Examples
    --------
    >>> import chemcat as cat
    >>> import chemcat.utils as u
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt

    >>> nlayers = 81
    >>> temperature = np.tile(1500.0, nlayers)
    >>> pressure = np.logspace(-8, 3, nlayers)
    >>> molecs = (
    >>>     'H2O CH4 CO CO2 NH3 N2 H2 HCN C2H2 C2H4 OH H He C N O '
    >>>     'e- H- H+ H2+ He+ '
    >>>     'Na Na- Na+ K K- K+ '
    >>>     'Si S SiO SiH4 H2S HS SO SO2 SiS'
    >>> ).split()

    >>> net = cat.Network(pressure, temperature, molecs)
    >>> vmr = net.thermochemical_equilibrium()
    >>> ax = u.plot_vmr(pressure, vmr, net.species, vmr_range=(1e-20,3))
    """
    species = list(species)

    if rect is not None:
        position = rect[0], rect[1], rect[2]-rect[0], rect[3]-rect[1]
    else:
        position = 0.09, 0.11, 0.79, 0.83

    if fignum == 316:
        fignum = 'Plotty McPltFace'

    # neutralized names:
    neutral_species = []
    for spec in species:
        neutral = spec.replace('-','').replace('+','')
        if neutral not in neutral_species:
            neutral_species.append(neutral)

    if colors is None:
        colors = resolve_colors(species)
    else:
        colors = {
            spec: color
            for spec,color in zip(neutral_species, colors)
        }

    dashes = {
        'neutral': (),
        'cation': (5.0, 1.5),
        'ion': (2.0, 0.75),
    }
    labels = ['neutral', 'cation', 'ion']
    ion_handles = [
        Line2D([], [], color='k', lw=1.5, dashes=dashes[label], label=label)
        for label in labels
    ]
    has_ions = np.any(['+' in spec or '-' in spec for spec in species])

    # Get on with the plot:
    if axis is None:
        fig = plt.figure(fignum, (8.5,5.0))
        plt.clf()
        ax = plt.subplot(111)
    else:
        ax = axis
    ax.set_position(position)
    ion_legend = ax.legend(
        handles=ion_handles,
        fontsize=np.clip(fontsize-3, 5, np.inf),
        loc='lower left', labelspacing=0.1, framealpha=0.6,
    )

    for name in species:
        charge = int('+' in name) - int('-' in name)
        charge_label = labels[charge]
        spec = name.replace('+','').replace('-','')
        if charge == 0 or spec not in species:
            label = name
        else:
            label = None
        ax.loglog(
            vmr[:,species.index(name)], pressure,
            label=label, lw=linewidth, color=colors[spec],
            dashes=dashes[charge_label],
        )

    ax.tick_params(
        which='both', right=True, top=True, direction='in',
        labelsize=fontsize-2,
    )
    ax.set_ylim(np.amax(pressure), np.amin(pressure))
    if vmr_range is not None:
        ax.set_xlim(vmr_range)
    ax.set_ylabel('Pressure (bar)', fontsize=fontsize)
    ax.set_xlabel('Volume mixing ratio', fontsize=fontsize)
    if has_ions:
        ax.add_artist(ion_legend)
    if title is not  None:
        ax.set_title(title, fontsize=fontsize)

    leg_args = {
        'loc': (1.01, 0.0),
        'fontsize': np.clip(fontsize-5, 5, np.inf),
        'ncol': 1,
        'columnspacing': 1.0,
        'labelspacing': 0.0,
    }
    ax.legend(**leg_args)

    # Matplotlib black magic:
    def on_draw(event):
        """This will be called once the figure is drawn"""
        ax = event.canvas.figure.vmr_axis
        legend = ax.get_legend()
        height_ratio = (
            legend.get_window_extent().height / ax.get_window_extent().height
        )
        ncols = int(height_ratio) + 1
        if ncols > 1:
            if fig.update_position:
                ax.set_position([0.09, 0.11, 0.7, 0.83])
            leg_args['ncol'] = ncols
            ax.legend(**leg_args)
            fig.canvas.draw()

    fig.vmr_axis = ax
    fig.update_position = axis is None and rect is None
    fig.canvas.mpl_connect('draw_event', on_draw)

    if savefig is not None:
        plt.savefig(savefig)
    ax.colors = colors
    return ax


def write_file(file, species, pressure, temperature, vmr):
    """
    Write pressure, temperature, and vmr values to file.

    Parameters
    ----------
    file: String
        Output file name.
    species: 1D string iterable
        Names of atmospheric species.
    pressure: 1D float iterable
        Atmospheric pressure profile (bar).
    temperature: 1D float iterable
        Atmospheric temperature profile (kelvin).
    vmr: 2D float iterable
        Atmospheric volume mixing ratios, of shape [nspecies, nlayers].

    Examples
    --------
    >>> import chemcat as cat
    >>> import numpy as np

    >>> nlayers = 81
    >>> temperature = np.tile(1200.0, nlayers)
    >>> pressure = np.logspace(-8, 3, nlayers)
    >>> molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH H He C N O'.split()
    >>> net = cat.Network(pressure, temperature, molecules)
    >>> vmr = net.thermochemical_equilibrium()

    >>> # Save results to file:
    >>> cat.utils.write_file(
    >>>     'chemcat_chemistry.dat', net.species, pressure, temperature, vmr,
    >>> )
    >>> # Read from file:
    >>> d = cat.utils.read_file('chemcat_chemistry.dat')
    """
    str_species = ' '.join(f'{spec:<15}' for spec in species)

    with open(file, 'w') as f:
        # Header:
        f.write(
            '# chemcat chemistry composition calculation '
            '(bar, K, volume mixing ratios)\n'
            f'# pressure   temperature  {str_species.rstrip()}\n\n'
        )
        # Per-layer data:
        for press, temp, vmr_layer in zip(pressure, temperature, vmr):
            str_vmr = ' '.join(f'{q:<15.8e}' for q in vmr_layer)
            f.write(f'{press:.8e}  {temp:8.2f}  {str_vmr}\n')


def read_file(file):
    r"""
    Read a chemcat file.

    Parameters
    ----------
    file: String
        Path to file to read.

    Returns
    -------
    species: 1D string list
        Names of atmospheric species.
    pressure: 1D float array
        Atmospheric pressure profile (bar).
    temperature: 1D float array
        Atmospheric temperature profile (kelvin).
    vmr: 2D float array
        Atmospheric volume mixing ratios, of shape [nspecies, nlayers].

    Examples
    --------
    >>> import chemcat.utils as u

    >>> # Continuing from example in u.write_file(),
    >>> # Read from file:
    >>> species, pressure, temperature, vmr = u.read_file(
    >>>     'chemcat_chemistry.dat')
    """
    with open(file, 'r') as f:
        _= f.readline()
        header = f.readline()
    species = header.split()[3:]

    data = np.loadtxt(file, unpack=True)
    pressure = data[0]
    temperature = data[1]
    vmr = data[2:].T

    return species, pressure, temperature, vmr



# Copyright (c) 2022 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'is_in',
    'read_thermo_build',
    'heat_func',
    'gibbs_func',
    'setup_network',
    'find_species',
]

import sys
from collections.abc import Iterable

from more_itertools import sliced
import numpy as np

from ..utils import ROOT
sys.path.append(f'{ROOT}chemcat/lib')
import _utils as u


def is_in(species, thermo_file=None):
    r"""
    Element-wise check whether species name exist in CEA database.
    Parameters
    ----------
    species: 1D iterable of strings
        Names of species to search in the database.
    thermo_file: String
        Optional ThermoBuild CEA database file path.
    Returns
    -------
    in_database: 1D bool array
        Flag whether each species is in the database.
    Examples
    --------
    >>> import chemcat.cea as cea
    >>> species = 'H2O (KOH)2 HO2 CO'.split()
    >>> in_cea = cea.is_in(species)
    >>> for spec, is_in in zip(species, in_cea):
    >>>     print(f'{spec:6s}  {is_in}')
    H2O     True
    (KOH)2  False
    HO2     True
    CO      True
    """
    if thermo_file is None:
        thermo_file = f'{ROOT}chemcat/data/thermo_build_cea.dat'

    with open(thermo_file, 'r') as f:
        lines = f.readlines()
    nlines = len(lines)

    all_species = []
    i = 0
    while i < nlines:
        line = lines[i]
        all_species.append(line[0:16].strip())
        line = lines[i+1]
        ndata_intervals = int(line[0:2])
        i += 2 + 3*ndata_intervals

    in_database = np.isin(species, all_species)
    return in_database


def read_thermo_build(species, thermo_file=None):
    """
    Read data from NASA's CEA thermoBuild file.
    https://cearun.grc.nasa.gov/ThermoBuild/index_ds.html

    Parameters
    ----------
    species: 1D iterable of string
        List of species names to extract their info.
    thermo_file: String
        Path to a file containing CEA ThermoBuild data.

    Returns
    -------
    thermo_data: Dict
        A dictionary containing the species thermal properties from the
        CEA database (one entry for each species):
        name: String
            Species name.
        stoich: Dict
            Stoichiometric value of the species as element-value pairs.
        a_coeffs: 2D float ndarray
            Polynomial coefficients to reproduce the heat capacity data.
        b_coeffs: 2D float ndarray
            Integration constants to obtain the enthalpy and entropy.
        t_coeffs: 1D float ndarray
            Temperature intervals of validity for each set of coefficients.

    Examples
    --------
    >>> import chemcat.cea as cea

    >>> # A simple HCNO network:
    >>> hcno_species = 'H2O CH4 CO CO2 NH3 HCN N2 H2 H He'.split()
    >>> hcno_thermo_data = cea.read_thermo_build(hcno_species)

    >>> # Network will all species from the database:
    >>> all_thermo_data = cea.read_thermo_build(species=None)
    """
    if thermo_file is None:
        thermo_file = f'{ROOT}chemcat/data/thermo_build_cea.dat'

    with open(thermo_file, 'r') as f:
        lines = f.readlines()
    nlines = len(lines)

    all_species = []
    loc_species = []
    i = 0
    ndata_max = 0
    while i < nlines:
        line = lines[i]
        all_species.append(line[0:16].strip())
        loc_species.append(i)
        line = lines[i+1]
        ndata_intervals = int(line[0:2])
        i += 2 + 3*ndata_intervals
        ndata_max = np.amax([ndata_max, ndata_intervals])

    if species is None:
        species = all_species

    thermo_data = []
    for species_name in species:
        if species_name not in all_species:
            print(f'{species_name} not in Thermo Build database')
            continue
        i = loc_species[all_species.index(species_name)]
        line = lines[i+1]
        ndata_intervals = int(line[0:2])
        # Chemical composition:
        stoich = {}
        for element in sliced(line[10:50], 8):
            if element[0:2].strip() == '':
                break
            element_name = element[0:2].strip().capitalize()
            if element_name == 'E':
                element_name = 'e'
            stoich[element_name] = float(element[2:])
        t_coeffs = np.zeros(ndata_max+1)
        a_coeffs = np.zeros((ndata_max,7))
        b_coeffs = np.zeros((ndata_max,2))
        for j in range(ndata_intervals):
            line = lines[i+3*j+2]
            t_coeffs[j:j+2] = line[0:22].split()
            line = lines[i+3*j+3].replace('D','E')
            a_coeffs[j,0:5] = list(sliced(line[0:80],16))
            line = lines[i+3*j+4].replace('D','E')
            a_coeffs[j,5:7] = list(sliced(line[0:32],16))
            b_coeffs[j] = list(sliced(line[48:80],16))
        thermo_data.append({
            'name': species_name,
            'stoich': stoich,
            'a_coeffs': a_coeffs,
            'b_coeffs': b_coeffs,
            't_coeffs': t_coeffs,
        })

    return thermo_data


def heat_func(a_coeffs, t_coeffs):
    """
    Generate a callable that evaluates the molar heat capacity
    at a given temperature array.

    Parameters
    ----------
    a_coeffs: 2D float ndarray
        Polynomial coefficients to reproduce the heat capacity data.
    t_coeffs: 1D float ndarray
        Temperature intervals of validity for each set of coefficients.

    Returns
    -------
    heat: Callable
        A function heat(temperature) that evaluates the molar heat
        capacity, cp(T)/R, for a given temperature input
        (which can be a single value or a 1D iterable).

    Examples
    --------
    >>> import chemcat.cea as cea

    >>> data = cea.read_thermo_build(['H2O'])[0]
    >>> heat = cea.heat_func(
    >>>     data['a_coeffs'], data['t_coeffs'])

    >>> print(heat(300.0))
    [4.04063805]
    >>> print(heat([300.0, 1000.0, 3000.0]))
    [4.04063805 4.96614188 6.8342561 ]
    """
    def heat(temperature):
        if not isinstance(temperature, Iterable):
            temperature = [temperature]
        temperature = np.array(temperature, np.double)

        heat_capacity = u.heat(
            temperature, a_coeffs, t_coeffs)
        return heat_capacity
    return heat


def gibbs_func(a_coeffs, b_coeffs, t_coeffs):
    """
    Generate a callable that evaluates the Gibbs free energy
    for a given temperature array.

    Parameters
    ----------
    a_coeffs: 2D float ndarray
        Polynomial coefficients to reproduce the heat capacity data.
    b_coeffs: 2D float ndarray
        Integration constants to obtain the enthalpy and entropy.
    t_coeffs: 1D float ndarray
        Temperature intervals of validity for each set of coefficients.

    Returns
    -------
    gibbs: Callable
        A function gibbs(temperature) that evaluates the Gibbs free
        energy, G(T)/RT, for a given temperature input (which can be
        a single value or a 1D iterable).

    Examples
    --------
    >>> import chemcat.cea as cea

    >>> data = cea.read_thermo_build(['H2O'])[0]
    >>> gibbs = cea.gibbs_func(
    >>>     data['a_coeffs'], data['b_coeffs'], data['t_coeffs'])

    >>> print(gibbs(300.0))
    [-119.66025955]
    >>> print(gibbs([300.0, 1000.0, 3000.0]))
    [-119.66025955  -53.94898416  -39.09425268]
    """
    def gibbs(temperature):
        if not isinstance(temperature, Iterable):
            temperature = [temperature]
        temperature = np.array(temperature, np.double)

        free_energy = u.gibbs(
            temperature, a_coeffs, b_coeffs, t_coeffs)
        return free_energy
    return gibbs


def setup_network(input_species):
    r"""
    Extract CEA thermal data for a requested chemical system.

    Parameters
    ----------
    species: 1D string iterable
        Species to search in the CEA data base.

    Returns
    -------
    species: 1D string array
        Species found in the CEA database (might differ from
        input_species if there are species not found on the database).
    heat_capacity: 1D list of callable objects
        Functions that evaluate the species's heat capacity (cp/R)
        at requested temperatures.
    gibbs_free_energy: 1D list of callable objects
        Functions that evaluate the species's Gibbs free energy (G/RT)
        at requested temperatures.
    stoich_data: List of dictionaries
        Stoichiometric data (as dictionary of element-value pairs) for
        a list of species.

    Examples
    --------
    >>> import chemcat.cea as cea

    >>> molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH H He C N O'.split()
    >>> species, heat_capacity, gibbs, stoich_data = \
    >>>     cea.setup_network(molecules)

    >>> for spec, stoich in zip(species, stoich_data):
    >>>     print(f'{spec:3s}:  {stoich}')
    H2O:  {'H': 2.0, 'O': 1.0}
    CH4:  {'C': 1.0, 'H': 4.0}
    CO :  {'C': 1.0, 'O': 1.0}
    CO2:  {'C': 1.0, 'O': 2.0}
    NH3:  {'N': 1.0, 'H': 3.0}
    N2 :  {'N': 2.0}
    H2 :  {'H': 2.0}
    HCN:  {'H': 1.0, 'C': 1.0, 'N': 1.0}
    OH :  {'O': 1.0, 'H': 1.0}
    H  :  {'H': 1.0}
    He :  {'He': 1.0}
    C  :  {'C': 1.0}
    N  :  {'N': 1.0}
    O  :  {'O': 1.0}
    """
    # Find which species exists in data base:
    thermo_data = read_thermo_build(input_species)

    cea_species = [data['name'] for data in thermo_data]
    idx_missing = np.array([
        spec not in cea_species for spec in input_species])
    if np.any(idx_missing):
        missing_species = np.array(input_species)[idx_missing]
        print(
            'These input species were not found in CEA database:\n'
            f'  {missing_species}')

    species = np.array(input_species)[~idx_missing]

    heat_capacity = []
    gibbs_free_energy = []
    stoich_data = []
    for data in thermo_data:
        heat_capacity.append(
            heat_func(data['a_coeffs'], data['t_coeffs']))
        gibbs_free_energy.append(
            gibbs_func(data['a_coeffs'], data['b_coeffs'], data['t_coeffs']))
        stoich_data.append(data['stoich'])

    return (
        species,
        heat_capacity,
        gibbs_free_energy,
        stoich_data,
    )


def find_species(elements, charge='neutral', num_atoms=None):
    """
    Find all CEA species that contain the specified properties
    (elements, charge, state).

    Parameters
    ----------
    elements: Dict or 1D string iterable
        Either:
        - A list of elements that must be present in the species, or
        - A dictionary of elements and their stoichiometric values.
    charge: String
        If 'neutral', limit the output only to neutrally charged species.
        If 'ion', limit the output only to charged species.
        Else, do not limit output.
    num_atoms: Integer
        Limit the number of atoms to the requested value.

    Returns
    -------
    species: 1D string array
        List of all species containing the required elements.

    Examples
    --------
    >>> import chemcat.cea as cea
    >>> # Get all sodium-bearing species:
    >>> species = cea.find_species(['Na'])
    >>> print(species)
    ['KNa' 'Na' 'NaCN' 'NaH' 'NaNO2' 'NaNO3' 'NaO' 'NaOH' 'Na2' 'Na2O' 'Na2O2'
     'Na2O2H2']

    >>> # Get species containing exactly two Na atoms:
    >>> species = cea.find_species({'Na':2})
    >>> print(species)
    ['Na2' 'Na2O' 'Na2O2' 'Na2O2H2']

    >>> # Species containing exactly two Na atoms and any amount of oxygen:
    >>> species = cea.find_species({'Na':2, 'O':None})
    >>> print(species)
    ['Na2O' 'Na2O2' 'Na2O2H2']

    >>> # Get all species containing sodium and oxygen (any amount):
    >>> species = cea.find_species(['Na', 'O'])
    >>> print(species)
    ['NaNO2' 'NaNO3' 'NaO' 'NaOH' 'Na2O' 'Na2O2' 'Na2O2H2']

    >>> # Get all hydrogen-ion species:
    >>> H_ions= cea.find_species(['H'], charge='ion')
    >>> print(H_ions)
    ['CH+' 'CH2OH+' 'H+' 'H-' 'HCO+' 'HD+' 'HO2-' 'H2+' 'H2-' 'H2O+' 'H3O+'
     'NH+' 'NH4+' 'NaOH+' 'OH+' 'OH-' 'MgOH+' 'SH-' 'SiH+']

    >>> # Only diatomic Na species:
    >>> diatomic = cea.find_species(['Na'], num_atoms=2, charge='all')
    >>> print(diatomic)
    ['KNa' 'NaH' 'NaO' 'Na2']
    """
    # Turn into dict if needed:
    if not isinstance(elements, dict):
        elements = {e:None for e in elements}

    with open(f'{ROOT}chemcat/data/thermo_build_cea.dat', 'r') as f:
        lines = f.readlines()
    nlines = len(lines)

    species = []
    i = 0
    while i < nlines:
        line = lines[i]
        name = line[0:16].strip()
        line = lines[i+1]
        ndata_intervals = int(line[0:2])
        i += 2 + 3*ndata_intervals

        stoich = {}
        for element in sliced(line[10:50], 8):
            if element[0:2].strip() == '':
                break
            element_name = element[0:2].strip().capitalize()
            if element_name == 'E':
                element_name = 'e'
            stoich[element_name] = float(element[2:])

        is_charged = 'e' in stoich
        if charge == 'neutral' and is_charged:
            continue
        elif charge == 'ion' and not is_charged:
            continue

        if num_atoms is not None:
            n_atoms = sum([val for key,val in stoich.items() if key!='e'])
            if n_atoms != num_atoms:
                continue

        # Request selected stoichiometric values:
        for key, val in elements.items():
            if key not in stoich or (val is not None and stoich[key] != val):
                break
        else:
            species.append(name)

    return np.array(species)




#!/usr/bin/env python
# coding: utf-8


# Chemistry Tutorial
# This tutorial shows how to compute chemistry with ``chemcat`` under different assumptions for temperature, metallicity, and custom elemental abundances.

# Let's start off importing the necessary modules:

import chemcat as cat
import chemcat.utils as u
import numpy as np


## Basic Run
# ``chemcat`` is an object-oriented code that keeps all the thermochemical properties in place.  In it most basic usage, all it requres is a pressure-temperature profile and a list of species.
# 
# This creates a chemical network assuming solar elemental composition and a simple CHNO compostion (taking values from Asplund et al. 2021):

nlayers = 81
temperature = np.tile(1200.0, nlayers)
pressure = np.logspace(-8, 3, nlayers)
molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH C2H2 C2H4 H He C N O'.split()

net = cat.Network(pressure, temperature, molecules)

# To compute thermochemical equilibrium abundances, run this method:
vmr = net.thermochemical_equilibrium()

# and plot the results:
title = f'T = {temperature[0]} K'
ax = u.plot_vmr(pressure, vmr, net.species, title=title, fignum=10)


### Evaluate at a different temperature
# Simply, provide a new temperature profile in the call:
temp_2800K = np.tile(2800, nlayers)
vmr = net.thermochemical_equilibrium(temperature=temp_2800K)

title = f'T = {temp_2800K[0]} K'
vmr_range = (1e-25, 3)
ax = u.plot_vmr(
    pressure, vmr, net.species,
    vmr_range=vmr_range,
    title=title,
    fignum=20,
)


### Object's data
# The ``net`` object contains all the physical data of the atmosphere, which you can query at any point:

print(f'Pressure profile (bar):\n{net.pressure}')
print(f'\nTemperature profile (kelvin):\n{net.temperature}')

print(f'\nElemental composition:\n{net.elements}')
print('\nElemental abundances (normalized to H_abundance==1.0)'
      f'\n(same order as in the array above):\n{net.element_rel_abundance}')
print(f'\nElemental composition source:\n{net.element_file}')

print(f'\n\nMolecular composition:\n{net.species}')
print(f'Source for each species thermochemical data:\n{net.provenance}')
print(
    '\nStoichiometric values matrix for each species:'
    '\n(order of values goes according to order of elements and species arrays)'
    f'\n{net.stoich_vals}')


## Custom compositions
# There are a few alternatives evaluate thermochemical equilibrium for different compositions

### Custom metallicity
# Use the `metallicity` argument to set a custom metallicity relative to solar (in dex units)

# E.g., solar metallicity (default):
net = cat.Network(pressure, temperature, molecules)
vmr_custom_solar = net.thermochemical_equilibrium(
    metallicity=0.0,
)

# e.g, 0.1x solar metallicity:
vmr_custom_sub_solar = net.thermochemical_equilibrium(
    metallicity=-1.0,
)

# e.g, 50x solar metallicity:
metallicity = np.log10(50.0)
vmr_custom_super_solar = net.thermochemical_equilibrium(
    metallicity=metallicity,
)


# And plot the results:
title = 'Custom sub solar metallicity (0.1x solar)'
ax = u.plot_vmr(
    pressure, vmr_custom_sub_solar, net.species,
    title=title,
    fignum=30,
)

title = 'Custom solar metallicity'
ax = u.plot_vmr(
    pressure, vmr_custom_solar, net.species,
    title=title,
    fignum=31,
)

title = 'Custom super solar metallicity (50x solar)'
ax = u.plot_vmr(
    pressure, vmr_custom_super_solar, net.species,
    title=title,
    fignum=32,
)


# Note that once you modify a physical property, it stays modified until you re-set it again:

print(f'[M/H] = {net.metallicity:.2f}')
print(f'Elemental abundances:\n{net.element_rel_abundance}')

# You can modify the properties directly:
net.metallicity = 0.0
print(f'\n[M/H] = {net.metallicity}')
print(f'Elemental abundances:\n{net.element_rel_abundance}')

# Or, you can modify the properties via a call:
# e.g, 50x solar:
vmr_custom_super_solar = net.thermochemical_equilibrium(
    metallicity=np.log10(50.0),
)
print(f'\n[M/H] = {net.metallicity:.2f}')
print(f'Elemental abundances:\n{net.element_rel_abundance}')


### Custom elemental abundances
# Use the `e_abundances` argument to set  custom elemental abundances as a dictionary: (values in dex units, relative to H=12.0):

# E.g., set carbon abundance to 10^-4
e_abundances = {
    'C': 8.0,
}
vmr_custom_abundance1 = net.thermochemical_equilibrium(
    metallicity=0.0,
    e_abundances=e_abundances,
)

# You can set values for as many elements as you want, e.g.:
e_abundances = {
    'C': 8.5,
    'N': 7.5,
    'O': 8.0,
}
vmr_custom_abundance2 = net.thermochemical_equilibrium(
    e_abundances=e_abundances,
)

# And plot the results:
title = 'Custom carbon abundance'
ax = u.plot_vmr(
    pressure, vmr_custom_abundance1, net.species,
    title=title,
    fignum=33,
)

title = 'Custom many-elements abundance (C/O>1.0)'
ax = u.plot_vmr(pressure, vmr_custom_abundance2, net.species,
    title=title,
    fignum=34,
)


### Elemental abundances relative to solar
# Use the `e_scale` argument to set custom elemental abundances as a dictionary (values in dex units, relative to solar):

# E.g., Set 10x solar carbon, 1x nitrogen, and 1x oxygen
e_scale = {
    'C': 1.0,
    'N': 0.0,
    'O': 0.0,
}

vmr_custom_scale_carbon = net.thermochemical_equilibrium(
    metallicity=0.0,
    e_abundances={},
    e_scale=e_scale,
)

# And plot the results:
title = 'Custom carbon abundance (scaled 10x solar), all other 1x solar'
ax = u.plot_vmr(
    pressure, vmr_custom_scale_carbon, net.species,
    title=title,
    fignum=35,
)


### Custom elemental ratios
# Use the `e_ratio` argument to set custom elemental abundances as a dictionary (values are **not** in dex units, relative to a second elemental abundance).  

# Note that the dictionary keys define the pair of elements separated by an underscore.
# Only the first element abundance is modified, and is scaled relative to the second element's abundance.

# E.g., set carbon abundance to 1.5x times the oxygen abundance:
e_ratio = {
    'C_O': 1.5,
}

vmr_custom_ratio = net.thermochemical_equilibrium(
    metallicity=0.0,
    e_abundances={},
    e_ratio=e_ratio,
)

# And plot the results:
title = 'Custom C/O abundance ratio (C/O = 1.5)'
ax = u.plot_vmr(pressure, vmr_custom_ratio, net.species, title=title, fignum=36)

# And you can combine all these properties at once:
# Custom metallicity, C/O ratio, and temperatures:

# Hot, sub-solar (0.1x), C/O < 1.0:
temp_2000K = np.tile(2000.0, nlayers)
metal_sub_solar = -1.0
e_ratio = {
    'C_O': 0.6,
}
vmr_custom1 = net.thermochemical_equilibrium(
    temperature=temp_2000K,
    metallicity=metal_sub_solar,
    e_ratio=e_ratio,
)

# Colder, super-solar (50x), and C/O > 1.0:
temp_1400K = np.tile(1400.0, nlayers)
metal_super_solar = np.log10(50.0)
e_ratio = {
    'C_O': 1.5,
}
vmr_custom2 = net.thermochemical_equilibrium(
    temperature=temp_1400K,
    metallicity=metal_super_solar,
    e_ratio=e_ratio,
)

# And plot the results:
vmr_range = (1e-20, 1)
title = 'Custom planet 1'
ax = u.plot_vmr(
    pressure, vmr_custom1, net.species,
    title=title, fignum=37, vmr_range=vmr_range,
)

title = 'Custom planet 2'
ax = u.plot_vmr(
    pressure, vmr_custom2, net.species,
    title=title, fignum=38, vmr_range=vmr_range,
)


## Charge conservation
# - Charge conservation is also available, ions are specified by species names followed by a ``+`` or a ``-``
# - Normally, you always need to include an electron in the mix: 'e-'

nlayers = 81
temp = 2000.0
temperature = np.tile(temp, nlayers)
pressure = np.logspace(-10, 3, nlayers)

molecules = (
    'H2O CH4 CO CO2 NH3 N2 H2 HCN C2H2 C2H4 OH H He C N O '
    'e- H- H+ H2+ He+ '
    'Na Na- Na+ K K- K+ '
    'Mg Mg+ Fe Fe+ '
    'Ti TiO TiO2 Ti+ TiO+ V VO VO2 V+ '
).split()

net = cat.Network(pressure, temperature, molecules)
vmr = net.thermochemical_equilibrium()
vmr_range = (1e-30, 3)
ax = u.plot_vmr(pressure, vmr, net.species, vmr_range=vmr_range, fignum=40)




# Copyright (c) 2022 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'Thermo_Prop',
    'Network',
]

import numpy as np

from . import janaf
from . import cea
from . import utils as u


class Thermo_Prop():
    """
    Descriptor objects that automate setting the elemental abundances.

    To understand this sorcery see:
    https://docs.python.org/3/howto/descriptor.html
    """
    def __set_name__(self, obj, name):
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        value = getattr(obj, self.private_name)
        return value

    def __set__(self, obj, value):
        setattr(obj, self.private_name, value)
        has_all_attributes = (
            hasattr(obj, 'metallicity')
            and hasattr(obj, 'e_abundances')
            and hasattr(obj, 'e_scale')
            and hasattr(obj, 'e_ratio')
        )
        if has_all_attributes:
            obj.element_rel_abundance = u.set_element_abundance(
                obj.elements,
                obj._base_composition,
                obj._base_dex_abundances,
                obj.metallicity, obj.e_abundances,
                obj.e_scale, obj.e_ratio,
            )


class Network(object):
    """A chemcat chemical network object."""
    metallicity = Thermo_Prop()
    e_abundances = Thermo_Prop()
    e_scale = Thermo_Prop()
    e_ratio = Thermo_Prop()

    def __init__(
        self, pressure, temperature, input_species,
        metallicity=0.0,
        e_abundances={},
        e_scale={},
        e_ratio={},
        e_source='asplund_2021',
        sources=['janaf', 'cea'],
    ):
        r"""
        Parameters
        ----------
        pressure: 1D float iterable
            Pressure profile (bar).
        temperature: 1D float iterable
            Temperature profile (Kelvin).
        input_species: 1D string iterable
            List of atomic, molecular, and ionic species to include in the
            chemical network.  Species not found in the database will be
            discarded.
        metallicity: Float
            Scaling factor for all elemental species except H and He.
            Dex units relative to the sun (e.g., solar is metallicity=0.0,
            10x solar is metallicity=1.0).
        e_abundances: Dictionary
            Custom elemental abundances.
            The dict contains the name of the element and their custom
            abundance in dex units relative to H=12.0.
            These values override metallicity.
        e_scale: Dictionary
            Custom elemental abundances scaled relative to solar values.
            The dict contains the name of the element and their custom
            scaling factor in dex units, e.g., for 2x solar carbon set
            e_scale = {'C': np.log10(2.0)}.
            This argument modifies the abundances on top of any custom
            metallicity and e_abundances.
        e_ratio: Dictionary
            Custom elemental abundances scaled relative to another element.
            The dict contains the pair of elements joined by an underscore
            and their ratio, e.g., for a C/O ratio of 0.8 set
            e_ratio = {'C_O': 0.8}.
            These values modify the abundances on top of any custom
            metallicity, e_abundances, and e_scale.
        e_source: String
            Source of elemental composition used as base composition.
            If value is 'asplund_2021' or 'asplund_2009', adopt the
            solar elemental composition values from Asplund+2009 or
            Asplund+2021, respectively.
            Else, assume e_source is a path to a custom elemental
            composition file.
        sources: List of strings
            Name of databases where to get the thermochemical properties
            (in order of priority).  Available options: 'janaf' or 'cea'.

        Examples
        --------
        >>> import chemcat as cat
        >>> import chemcat.utils as u
        >>> import numpy as np

        >>> nlayers = 81
        >>> temperature = np.tile(1200.0, nlayers)
        >>> pressure = np.logspace(-8, 3, nlayers)
        >>> molecules = (
        >>>     'H2O CH4 CO CO2 NH3 N2 H2 HCN OH C2H2 C2H4 H He C N O').split()
        >>> net = cat.Network(pressure, temperature, molecules)

        >>> # Compute abundances in thermochemical equilibrium:
        >>> vmr = net.thermochemical_equilibrium()

        >>> # See results:
        >>> ax = u.plot_vmr(pressure, vmr, net.species, vmr_range=(1e-30,2))

        >>> # Compute heat capacity:
        >>> cp = net.heat_capacity()
        >>> print(f'Heat capacity (cp/R):\n{cp[0]}')
        Heat capacity (cp/R):
        [5.26408044 9.48143057 4.11030773 6.77638503 7.34238673 4.05594463
         3.72748083 6.3275286  3.79892261 2.49998117 2.49998117 2.50082308
         2.49998117 2.51092596]
        """
        self.pressure = pressure
        self.temperature = temperature
        self.input_species = input_species

        dealiased_species = u.de_aliasing(input_species, sources)

        # Check and remove duplicates:
        is_unique = np.zeros(len(dealiased_species), bool)
        for i,spec in enumerate(dealiased_species):
            if spec not in np.array(dealiased_species)[is_unique]:
                is_unique[i] = True
        if not np.all(is_unique):
            duplicates = np.array(input_species)[~is_unique]
            print(
                'These species are duplicates of others in input (and will '
                f'be removed):\n  {duplicates}'
            )
            input_species = np.array(input_species)[is_unique]
            dealiased_species = np.array(dealiased_species)[is_unique]

        # Check species exist in the database(s):
        source_names = u.resolve_sources(dealiased_species, sources)
        idx_valid = source_names != None
        self.provenance = source_names[idx_valid]
        self.species = np.array(input_species)[idx_valid]

        if None in source_names:
            missing_species = np.array(input_species)[~idx_valid]
            print(
                'These input species were not found in any database:'
                f'\n  {missing_species}'
            )

        nspecies = len(self.species)
        self._heat_capacity = np.zeros(nspecies, object)
        self._gibbs_free_energy = np.zeros(nspecies, object)
        stoich_data = np.tile(None, nspecies)

        for source in np.unique(self.provenance):
            if source == 'janaf':
                setup_network = janaf.setup_network
            if source == 'cea':
                setup_network = cea.setup_network
            species = np.array(dealiased_species)[source_names==source]
            network_data = setup_network(species)

            idx_db = self.provenance == source
            self._heat_capacity[idx_db] = network_data[1]
            self._gibbs_free_energy[idx_db] = network_data[2]
            stoich_data[idx_db] = network_data[3]

        self.elements, self.stoich_vals = u.stoich_matrix(stoich_data)

        # Setup input elemental composition:
        if e_source == 'asplund_2009' or e_source == 'asplund_2021':
            e_source = f'{u.ROOT}chemcat/data/{e_source}_solar_abundances.dat'

        self.element_file = e_source
        base_data = u.read_elemental(self.element_file)
        self._base_composition = base_data[0]
        self._base_dex_abundances = base_data[1]

        self.metallicity = metallicity
        self.e_abundances = e_abundances
        self.e_scale = e_scale
        self.e_ratio = e_ratio

        self.element_rel_abundance = u.set_element_abundance(
            self.elements,
            self._base_composition,
            self._base_dex_abundances,
            self.metallicity,
            self.e_abundances,
            self.e_scale,
            self.e_ratio,
        )


    def heat_capacity(self, temperature=None):
        """
        Evaluate the heat capacity of each species in the network
        at the given temperature (default to self.temperature if needed).
        """
        if temperature is None:
            temperature = self.temperature
        return u.thermo_eval(temperature, self._heat_capacity)


    def gibbs_free_energy(self, temperature=None):
        """
        Evaluate the Gibbs free energy of each species in the network
        at the given temperature (default to self.temperature if needed).
        """
        if temperature is None:
            temperature = self.temperature
        return u.thermo_eval(temperature, self._gibbs_free_energy)


    def thermochemical_equilibrium(
        self,
        temperature=None,
        metallicity=None,
        e_abundances=None,
        e_scale=None,
        e_ratio=None,
        savefile=None,
    ):
        """
        Compute thermochemical-equilibrium abundances, update the
        atmospheric properties according to any non-None argument.

        Parameters
        ----------
        temperature: 1D float iterable
            Temperature profile (Kelvin).
            Must have same number of layers as self.pressure.
        metallicity: Float
            Scaling factor for all elemental species except H and He.
            Dex units relative to the sun (e.g., solar is metallicity=0.0,
            10x solar is metallicity=1.0).
        e_abundances: Dictionary
            Elemental abundances for custom species set as
            {element: abundance} pairs in dex units relative to H=12.0.
            These values override metallicity.
        e_scale: Dictionary
            Custom elemental abundances scaled relative to solar values.
            The dict contains the name of the element and their custom
            scaling factor in dex units, e.g., for 2x solar carbon set
            e_scale = {'C': np.log10(2.0)}.
            This argument modifies the abundances on top of any custom
            metallicity and e_abundances.
        e_ratio: Dictionary
            Custom elemental abundances scaled relative to another element.
            The dict contains the pair of elements joined by an underscore
            and their ratio, e.g., for a C/O ratio of 0.8 set
            e_ratio = {'C_O': 0.8}.
            These values modify the abundances on top of any custom
            metallicity, e_abundances, and e_scale.
        savefile: String
            If not None, store vmr outputs to given file path.

        Returns
        -------
        vmr: 2D float array
            Species volume mixing ratios in thermochemical equilibrium
            of shape [nlayers, nspecies].
        """
        if temperature is not None:
            if np.shape(temperature) != np.shape(self.pressure):
                raise ValueError(
                    'Temperature profile does not match size of pressure '
                    f'profile ({len(self.pressure)} layers)'
                )
            self.temperature = temperature
        if metallicity is not None:
            self.metallicity = metallicity
        if e_abundances is not None:
            self.e_abundances = e_abundances
        if e_scale is not None:
            self.e_scale = e_scale
        if e_ratio is not None:
            self.e_ratio = e_ratio

        self.vmr = u.thermochemical_equilibrium(
            self.pressure,
            self.temperature,
            self.element_rel_abundance,
            self.stoich_vals,
            self._gibbs_free_energy,
        )

        if savefile is not None:
            u.write_file(
                savefile,
                self.species, self.pressure, self.temperature,
                self.vmr,
            )

        return self.vmr




# Copyright (c) 2022 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

import os

import numpy as np
import pytest

import chemcat as cat
import chemcat.utils as u
import chemcat.janaf as janaf

from conftest import *


# Ddefault values:
element_file = f'{u.ROOT}chemcat/data/asplund_2021_solar_abundances.dat'


def test_thermo_eval_heat_capacity_single_temp():
    molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH H He C N O'.split()
    janaf_data = janaf.setup_network(molecules)
    heat_capacity = janaf_data[1]
    temperature = 1500.0
    cp = u.thermo_eval(temperature, heat_capacity)

    expected_cp = np.array([
        5.6636252 , 10.41029396,  4.23563153,  7.02137982,  8.00580904,
        4.19064967,  3.88455652,  6.65454913,  3.95900511,  2.49998117,
        2.49998117,  2.5033488 ,  2.49998117,  2.50707724])
    np.testing.assert_allclose(cp, expected_cp)


def test_thermo_eval_heat_capacity_temp_array():
    molecules = 'H2O CH4 CO C He'.split()
    janaf_data = janaf.setup_network(molecules)
    heat_capacity = janaf_data[1]
    temperatures = np.arange(100.0, 4501.0, 200.0)
    cp = u.thermo_eval(temperatures, heat_capacity)

    np.testing.assert_allclose(cp, expected_cp)


def test_thermo_eval_gibbs_free_energy_temp_array():
    molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH H He C N O'.split()
    janaf_data = janaf.setup_network(molecules)
    gibbs_funcs = janaf_data[2]
    temperatures = np.arange(100.0, 4101.0, 500.0)
    gibbs = u.thermo_eval(temperatures, gibbs_funcs)

    np.testing.assert_allclose(gibbs, expected_gibbs)


def test_stoich_matrix_neutrals():
    stoich_data = [
        {'H': 2.0, 'O': 1.0},
        {'C': 1.0, 'H': 4.0},
        {'C': 1.0, 'O': 2.0},
        {'H': 2.0},
        {'H': 1.0},
        {'He': 1.0},
    ]
    expected_stoich = np.array([
        [0, 2, 0, 1],
        [1, 4, 0, 0],
        [1, 0, 0, 2],
        [0, 2, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
    ])

    elements, stoich_matrix = u.stoich_matrix(stoich_data)
    np.testing.assert_equal(elements, np.array(['C', 'H', 'He', 'O']))
    np.testing.assert_equal(stoich_matrix, expected_stoich)


def test_stoich_matrix_ions():
    stoich_data = [
        {'H': 2.0, 'O': 1.0},
        {'C': 1.0, 'H': 4.0},
        {'C': 1.0, 'O': 2.0},
        {'H': 2.0},
        {'H': 1.0, 'e': -1.0},
        {'He': 1.0},
        {'e': 1.0},
    ]
    expected_stoich = np.array([
        [ 0,  2,  0,  1,  0],
        [ 1,  4,  0,  0,  0],
        [ 1,  0,  0,  2,  0],
        [ 0,  2,  0,  0,  0],
        [ 0,  1,  0,  0, -1],
        [ 0,  0,  1,  0,  0],
        [ 0,  0,  0,  0,  1],
    ])

    elements, stoich_matrix = u.stoich_matrix(stoich_data)
    np.testing.assert_equal(elements, np.array(['C', 'H', 'He', 'O', 'e']))
    np.testing.assert_equal(stoich_matrix, expected_stoich)


@pytest.mark.parametrize('sources', ('janaf', ['janaf']))
def test_de_aliasing_janaf_only(sources):
    input_species = ['H2O', 'C2H2', 'HO2', 'CO']
    output_species = u.de_aliasing(input_species, sources)
    assert output_species == ['H2O', 'C2H2', 'HOO', 'CO']


@pytest.mark.parametrize('sources', ('cea', ['cea']))
def test_de_aliasing_cea_only(sources):
    input_species = ['H2O', 'C2H2', 'HO2', 'CO']
    output_species = u.de_aliasing(input_species, sources)
    assert output_species == ['H2O', 'C2H2,acetylene', 'HO2', 'CO']


def test_de_aliasing_janaf_cea():
    input_species = ['H2O', 'C2H2', 'HO2', 'CO']
    sources = ('janaf', 'cea')
    output_species = u.de_aliasing(input_species, sources)
    assert output_species == ['H2O', 'C2H2', 'HOO', 'CO']


def test_de_aliasing_not_found():
    input_species = ['H2O', 'C4H2', 'HO2', 'CO']
    sources = 'janaf'
    output_species = u.de_aliasing(input_species, sources)
    assert output_species == ['H2O', 'C4H2', 'HOO', 'CO']


def test_de_aliasing_default_cea():
    input_species = ['H2O', 'C4H2', 'HO2', 'CO']
    sources = ('janaf', 'cea')
    output_species = u.de_aliasing(input_species, sources)
    assert output_species == ['H2O', 'C4H2,butadiyne', 'HOO', 'CO']


def test_resolve_sources_with_missing_species():
    species = 'H2O CO (KOH)2 HO2'.split()
    sources = u.resolve_sources(species, sources=['cea'])
    assert list(sources) == ['cea', 'cea', None, 'cea']


def test_resolve_sources_cea_priority():
    species = 'H2O CO (KOH)2 HO2'.split()
    sources = u.resolve_sources(species, sources=['cea', 'janaf'])
    assert list(sources) == ['cea', 'cea', 'janaf', 'cea']


def test_resolve_sources_janaf_priority():
    species = 'H2O CO (KOH)2 HO2'.split()
    sources = u.resolve_sources(species, sources=['janaf', 'cea'])
    assert list(sources) == ['janaf', 'janaf', 'janaf', 'cea']


@pytest.mark.parametrize('sources', ('cea', ['cea']))
def test_resolve_sources_list_or_string(sources):
    species = 'H2O CO (KOH)2 HO2'.split()
    source_names = u.resolve_sources(species, sources=['cea'])
    assert list(source_names) == ['cea', 'cea', None, 'cea']


@pytest.mark.parametrize('sun',
    [
        'asplund_2009_solar_abundances.dat',
        'asplund_2021_solar_abundances.dat',
    ])
def test_read_elemental(sun):
    elements, dex = u.read_elemental(f'{u.ROOT}chemcat/data/{sun}')

    expected_elements_asplund = (
        'D   H   He  Li  Be  B   C   N   O   F   Ne  Na '
        'Mg  Al  Si  P   S   Cl  Ar  K   Ca  Sc  Ti  V '
        'Cr  Mn  Fe  Co  Ni  Cu  Zn  Ga  Ge  As  Se '
        'Br  Kr  Rb  Sr  Y   Zr  Nb  Mo  Ru  Rh  Pd '
        'Ag  Cd  In  Sn  Sb  Te  I   Xe  Cs  Ba  La '
        'Ce  Pr  Nd  Sm  Eu  Gd  Tb  Dy  Ho  Er  Tm '
        'Yb  Lu  Hf  Ta  W   Re  Os  Ir  Pt  Au  Hg '
        'Tl  Pb  Bi  Th  U').split()
    if '2009' in sun:
        expected_dex_asplund = expected_dex_asplund_2009
    elif '2021' in sun:
        expected_dex_asplund = expected_dex_asplund_2021

    assert len(elements) == 84
    for element, expected_element in zip(elements, expected_elements_asplund):
        assert element == expected_element
    np.testing.assert_allclose(dex[dex>0], expected_dex_asplund)


def test_set_element_abundance_solar():
    sun_elements, sun_dex = u.read_elemental(element_file)
    elements = 'H He C N O'.split()
    e_abundances = u.set_element_abundance(
        elements, sun_elements, sun_dex)
    expected_abundance = np.array([
        1.0, 8.20351544e-02, 2.88403150e-04, 6.76082975e-05, 4.89778819e-04,
    ])
    np.testing.assert_allclose(e_abundances, expected_abundance)

def test_set_element_abundance_metallicity():
    sun_elements, sun_dex = u.read_elemental(element_file)
    elements = 'H He C N O'.split()
    e_abundances = u.set_element_abundance(
        elements, sun_elements, sun_dex, metallicity=0.5)
    expected_abundance = np.array([
        1.0, 8.20351544e-02, 9.12010839e-04, 2.13796209e-04, 1.54881662e-03,
    ])
    np.testing.assert_allclose(e_abundances, expected_abundance)


def test_set_element_abundance_custom_element():
    sun_elements, sun_dex = u.read_elemental(element_file)
    elements = 'H He C N O'.split()
    e_abundances = u.set_element_abundance(
        elements, sun_elements, sun_dex, e_abundances={'C': 8.8})
    expected_abundance = np.array([
        1.0, 8.20351544e-02, 6.30957344e-04, 6.76082975e-05, 4.89778819e-04,
    ])
    np.testing.assert_allclose(e_abundances, expected_abundance)


def test_set_element_abundance_custom_e_scale():
    sun_elements, sun_dex = u.read_elemental(element_file)
    elements = 'H He C N O'.split()
    e_abundances = u.set_element_abundance(
        elements, sun_elements, sun_dex, e_scale={'C': np.log10(3.0)})
    expected_abundance = np.array([
        1.0, 8.20351544e-02, 8.65209451e-04, 6.76082975e-05, 4.89778819e-04,
    ])
    np.testing.assert_allclose(e_abundances, expected_abundance)


def test_set_element_abundance_custom_e_ratio():
    sun_elements, sun_dex = u.read_elemental(element_file)
    elements = 'H He C N O'.split()
    e_abundances = u.set_element_abundance(
        elements, sun_elements, sun_dex, e_ratio={'C_O': 0.6},
    )
    expected_abundance = np.array([
        1.0, 8.20351544e-02, 2.93867292e-04, 6.76082975e-05, 4.89778819e-04
    ])
    np.testing.assert_allclose(e_abundances, expected_abundance)


def test_write_file_read_file(tmpdir):
    atm_file = "atmfile.dat"
    atm = f'{tmpdir}/{atm_file}'

    nlayers = 11
    temperature = np.tile(1200.0, nlayers)
    pressure = np.logspace(-8, 3, nlayers)
    molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH H He C N O'.split()
    net = cat.Network(pressure, temperature, molecules)
    vmr = net.thermochemical_equilibrium()

    u.write_file(atm, net.species, pressure, temperature, vmr)
    assert atm_file in os.listdir(str(tmpdir))

    # Now, open file and check values:
    species, pressure, temperature, read_vmr = u.read_file(atm)
    np.testing.assert_equal(species, molecules)
    np.testing.assert_allclose(pressure, pressure)
    np.testing.assert_allclose(temperature, temperature)
    np.testing.assert_allclose(read_vmr, vmr)



def test_write_file_read_file_tiny_abundances(tmpdir):
    atm_file = "atmfile.dat"
    atm = f'{tmpdir}/{atm_file}'

    nlayers = 11
    temperature = np.tile(1200.0, nlayers)
    pressure = np.logspace(-8, 3, nlayers)
    molecules = 'H2O CH4 CO CO2 NH3 N2 H2 HCN OH H He C N O'.split()
    net = cat.Network(pressure, temperature, molecules)
    vmr = net.thermochemical_equilibrium()
    # Very small values change the print format:
    vmr[0:0] = 1.0e-200
    vmr[:,3] = 2.0e-200
    vmr[3,:] = 3.0e-200

    u.write_file(atm, net.species, pressure, temperature, vmr)
    assert atm_file in os.listdir(str(tmpdir))
    # Now, open file and check values:
    species, pressure, temperature, read_vmr = u.read_file(atm)
    np.testing.assert_allclose(pressure, pressure)
    np.testing.assert_allclose(temperature, temperature)
    np.testing.assert_equal(species, molecules)
    np.testing.assert_allclose(read_vmr, vmr)


def test_resolve_colors_defaults():
    species = 'H He C H2 CH4 CO CO2'.split()
    colors = u.resolve_colors(species)
    assert colors['H'] == 'blue'
    assert colors['He'] == 'olive'
    assert colors['C'] == 'coral'
    assert colors['H2'] == 'deepskyblue'
    assert colors['CH4'] == 'darkorange'
    assert colors['CO'] == 'limegreen'
    assert colors['CO2'] == 'red'


def test_resolve_colors_ions():
    species = 'H He C H2 CH4 CO CO2 e- H+ H- H3+'.split()
    colors = u.resolve_colors(species)
    # Ions not really counted as extra species-color pairs:
    assert 'H+' not in colors
    assert 'H-' not in colors
    # But neutral version of ions exist indeed:
    assert colors['e'] == 'darkgreen'
    assert colors['H3'] == 'royalblue'

    assert colors['H'] == 'blue'
    assert colors['He'] == 'olive'
    assert colors['C'] == 'coral'
    assert colors['H2'] == 'deepskyblue'
    assert colors['CH4'] == 'darkorange'
    assert colors['CO'] == 'limegreen'
    assert colors['CO2'] == 'red'


def test_resolve_colors_custom_list():
    species = 'H He C H2 CH4 CO CO2'.split()
    color_list = 'red green blue black orange brown magenta'.split()
    colors = u.resolve_colors(species, color_list=color_list)
    # If only color_list is input, do not default color_dict:
    assert colors['H'] == 'red'
    assert colors['He'] == 'green'
    assert colors['C'] == 'blue'
    assert colors['H2'] == 'black'
    assert colors['CH4'] == 'orange'
    assert colors['CO'] == 'brown'
    assert colors['CO2'] == 'magenta'


def test_resolve_colors_more_species_than_colors():
    species = 'H He C H2 CH4 CO CO2'.split()
    color_list = 'red green blue black'.split()
    colors = u.resolve_colors(species, color_list=color_list)
    assert colors['H'] == 'red'
    assert colors['He'] == 'green'
    assert colors['C'] == 'blue'
    assert colors['H2'] == 'black'
    # Colors cycle over remaining species:
    assert colors['CH4'] == 'red'
    assert colors['CO'] == 'green'
    assert colors['CO2'] == 'blue'


def test_resolve_colors_not_all_colors_in_dict():
    species = 'H He C H2 CH4 CO CO2'.split()
    color_dict = {
        'H': 'red',
        'He': 'green',
        'H2': 'blue',
    }
    colors = u.resolve_colors(species, color_dict=color_dict)
    # Colors as in input color_dict:
    assert colors['H'] == 'red'
    assert colors['He'] == 'green'
    assert colors['H2'] == 'blue'
    # Remaining species get assigned a color by default from u.COLORS:
    assert colors['C'] == 'royalblue'
    assert colors['CH4'] == 'darkorange'
    assert colors['CO'] == 'darkgreen'
    assert colors['CO2'] == 'magenta'


def test_plot_vmr():
    # Test that it doesn't break ... (most I can do)
    nlayers = 81
    temp = 1700.0
    temperature = np.tile(temp, nlayers)
    pressure = np.logspace(-10, 3, nlayers)

    species = (
        'H2O CH4 CO CO2 NH3 N2 H2 HCN C2H2 C2H4 OH H He C N O '
        'e- H- H+ H2+ He+ '
        'Na Na- Na+ K K- K+ '
        'Ti TiO TiO2 Ti+ TiO+ V VO VO2 V+ '
        'Si S SiO SiH4 H2S HS SO SO2 SiS'
    ).split()

    net = cat.Network(pressure, temperature, species)
    vmr = net.thermochemical_equilibrium()
    ax = u.plot_vmr(pressure, vmr, net.species)




#!/usr/bin/env python
# coding: utf-8

# Databases Tutorial
# This tutorial shows how to compute chemistry with ``chemcat`` under different assumptions for temperature, metallicity, and custom elemental abundances.
# 
# Let's start off importing the necessary modules and create an utility function to plot the results:


import chemcat.cea as cea
import chemcat.janaf as janaf


## Look up by Atom
# To search for species containing specific atoms call the `find_species()` function with a list of the atomic species requested:

# Get all sodium-bearing species:
janaf_salts = janaf.find_species(['Na'])
print(f'JANAF salts with any amount of sodium:\n{janaf_salts}')

cea_salts = cea.find_species(['Na'])
print(f'\nCEA salts with any amount of sodium:\n{cea_salts}')


# Get all sodium- and oxigen-bearing species:
janaf_sodium_oxide = janaf.find_species(['Na', 'O'])
print(
    'JANAF salts with any amount of sodium and oxygen:'
    f'\n{janaf_sodium_oxide}'
)

cea_sodium_oxide = cea.find_species(['Na', 'O'])
print(
    '\nCEA salts with any amount of sodium and oxigen:'
    f'\n{cea_sodium_oxide}'
)


## Look up by Atomic Number
# To search for species containing specific atoms call the `find_species()` function with a dict of the atomic species requested and their atomic values:

# Get species containing exactly two Na atoms:
janaf_salts1 = janaf.find_species({'Na':1})
print(f'JANAF salts with one sodium atom:\n{janaf_salts1}')

cea_salts1 = cea.find_species({'Na':1})
print(f'\nCEA salts with one sodium atom:\n{cea_salts1}')


# Get species containing exactly two Na atoms:
janaf_salts2 = janaf.find_species({'Na':2})
print(f'JANAF salts with two sodium atoms:\n{janaf_salts2}')

cea_salts2 = cea.find_species({'Na':2})
print(f'\nCEA salts with two sodium atoms:\n{cea_salts2}')


# Use `None` as atomic number to request any non-zero amount of atoms for a given species:

# Species containing exactly two carbon atoms and any amount of hydrogen:
janaf_ethane = janaf.find_species({'H':None, 'C':2})
print(f'JANAF ethane hydrocarbons:\n{janaf_ethane}')

cea_ethane = cea.find_species({'H':None, 'C':2})
print(f'\nCEA ethane hydrocarbons:\n{cea_ethane}')


## Look up by Number of Total Atoms
# To search for species containing a specific number of atoms call the `find_species()` function with the `num_atoms` argument:

# Species containing exactly two atoms and at least one oxygen:
janaf_diatomic_monoxides = janaf.find_species(['O'], num_atoms=2)
print(f'JANAF diatomic monoxides (and O2):\n{janaf_diatomic_monoxides}')

cea_diatomic_monoxides = cea.find_species(['O'], num_atoms=2)
print(f'\nCEA diatomic monoxides (and O2):\n{cea_diatomic_monoxides}')


## Look up Ionic Species
# So far we have seen only neutral species, if we want to look up for charged species we need to call the `find_species()` function with the `charge` argument:

# Look up sulfuric species in JANAF:
janaf_sulfur = janaf.find_species(['S'], charge='all')
print(f'All JANAF sulfuric species:\n{janaf_sulfur}')

janaf_sulfur_neutral = janaf.find_species(['S'], charge='neutral')
print(f'\nNeutral JANAF sulfuric species:\n{janaf_sulfur_neutral}')

janaf_sulfur_ions = janaf.find_species(['S'], charge='ion')
print(f'\nCharged JANAF sulfuric species:\n{janaf_sulfur_ions}')


# Look up sulfuric species in CEA:
cea_sulfur = cea.find_species(['S'], charge='all')
print(f'All CEA sulfuric species:\n{cea_sulfur}')

cea_sulfur_neutral = cea.find_species(['S'], charge='neutral')
print(f'\nNeutral CEA sulfuric species:\n{cea_sulfur_neutral}')

cea_sulfur_ions = cea.find_species(['S'], charge='ion')
print(f'\nCharged CEA sulfuric species:\n{cea_sulfur_ions}')




# Copyright (c) 2022 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

import os
import re
import setuptools
from setuptools import setup, Extension

from numpy import get_include


def get_version(package):
    """Return package version as listed in __version__ in version.py"""
    path = os.path.join(os.path.dirname(__file__), package, 'version.py')
    with open(path, 'rb') as f:
        init_py = f.read().decode('utf-8')
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


srcdir = 'src_c/'  # C-code source folder
incdir = 'src_c/include/'  # Include filder with header files

cfiles = os.listdir(srcdir)
cfiles = list(filter(lambda x: re.search('.+[.]c$', x), cfiles))
cfiles = list(filter(lambda x: not re.search('[.#].+[.]c$', x), cfiles))

inc = [get_include(), incdir]
eca = ['-ffast-math']
ela = []

extensions = [
    Extension(
        'chemcat.lib.' + cfile.rstrip('.c'),
        sources=[f'{srcdir}{cfile}'],
        include_dirs=inc,
        extra_compile_args=eca,
        extra_link_args=ela,
    )
    for cfile in cfiles
]

package_data = {
    'chemcat': [
        'data/*',
        'data/janaf/*',
    ]
}

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name = 'chemcat',
    version = get_version('chemcat'),
    author = 'Jasmina Blecic and Patricio Cubillos',
    author_email = 'patricio.cubillos@oeaw.ac.at',
    url = 'https://github.com/atmolib/chemcat',
    packages = setuptools.find_packages(),
    package_data = package_data,
    install_requires = [
        'numpy>=1.19.1',
        'scipy>=1.5.2',
        'matplotlib>=3.3.4',
        'more-itertools>=8.4.0',
        ],
    tests_require = [
        'pytest>=3.9',
        ],
    license = 'GPLv2',
    description = 'Chemistry Calculator for Atmospheres',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    include_dirs = inc,
    ext_modules = extensions,
)




# Copyright (c) 2022 Blecic and Cubillos
# chemcat is open-source software under the GPL-2.0 license (see LICENSE)

from .cea import *

__all__ = cea.__all__


# Clean up top-level namespace--delete everything that isn't in __all__
# or is a magic attribute, and that isn't a submodule of this package
for varname in dir():
    if not ((varname.startswith('__') and varname.endswith('__')) or
            varname in __all__):
        del locals()[varname]
del(varname)
